import json
import re
from datetime import datetime, timedelta
from typing import Dict, List, Union

from marshmallow import (
    RAISE,
    Schema,
    ValidationError,
    fields,
    post_dump,
    post_load,
    validate,
    validates,
    validates_schema,
)

from cognite.model_hosting._utils import calculate_windows, granularity_to_ms, time_interval_to_ms, timestamp_to_ms
from cognite.model_hosting.data_spec.exceptions import SpecValidationError

INVALID_AGGREGATE_FUNCTIONS = ["avg", "cv", "dv", "int", "step", "tv"]


class _BaseSpec:
    _schema = None  # Set by subclass

    def dump(self):
        """Dumps the data spec into a Python data structure.

        Raises:
            SpecValidationError: If the spec is not valid.

        Returns:
            Dict: The data spec as a Python data structure.
        """
        try:
            dumped = self._schema.dump(self)
        except ValidationError as e:
            raise SpecValidationError(e.messages) from e
        errors = self._schema.validate(dumped)
        if errors:
            raise SpecValidationError(errors)

        return dumped

    def validate(self):
        """Checks whether or not the data spec is valid.

        Raises:
            SpecValidationError: If the spec is not valid.
        """
        self.dump()

    @classmethod
    def load(cls, data):
        """Loads the data from a Python data structure.

        Raises:
            SpecValidationError: If the spec is not valid.

        Returns:
            The data spec object.
        """
        try:
            return cls._schema.load(data)
        except ValidationError as e:
            raise SpecValidationError(e.messages) from e

    def to_json(self):
        """Returns a json representation of the data spec.

        Raises:
            SpecValidationError: If the spec is not valid.

        Returns:
            str: The json representation of the data spec.
        """
        return json.dumps(self.dump(), indent=4, sort_keys=True)

    @classmethod
    def from_json(cls, s: str):
        """Loads the data spec from a json representation.

        Raises:
            SpecValidationError: If the spec is not valid.

        Returns:
            The data spec object.
        """
        return cls.load(json.loads(s))

    def copy(self):
        """Returns a copy of the data spec.

        Raises:
            SpecValidationError: If the spec is not valid.
        """
        return self.from_json(self.to_json())

    def __str__(self):
        return self.to_json()

    def __eq__(self, other):
        if type(self) == type(other):
            return self.__dict__ == other.__dict__
        else:
            return False


class TimeSeriesSpec(_BaseSpec):
    """Creates a time series spec.

    If the granularity and aggregate parameters are omitted, the TimeSeriesSpec specifies raw data.

    Args:
        id (int): The id of the time series.
        start (Union[str, int, datetime]): The (inclusive) start of the time series. Can be either milliseconds since epoch,
        time-ago format (e.g. "1d-ago"), or a datetime object.
        end (Union[str, int, datetime]): The (exclusive) end of the time series. Same format as start. Can also be set to "now".
        aggregate (str, optional): The aggregate function to apply to the time series.
        granularity (str, optional): Granularity of the datapoints. e.g. "1m", "2h", or "3d".
        include_outside_points (bool): Whether or not to include the first point before and after start and end. Can
                                        only be used with raw data.
    """

    def __init__(
        self,
        id: int,
        start: Union[int, str, datetime],
        end: Union[int, str, datetime],
        aggregate: str = None,
        granularity: str = None,
        include_outside_points: bool = None,
    ):
        self.id = id
        self.start = timestamp_to_ms(start)
        self.end = timestamp_to_ms(end)
        self.aggregate = aggregate
        self.granularity = granularity
        self.include_outside_points = include_outside_points


class FileSpec(_BaseSpec):
    """Creates a file spec.

    Args:
        id (int): The id of the file.
    """

    def __init__(self, id: int):
        self.id = id


class DataSpec(_BaseSpec):
    """Creates a DataSpec.

    This object collects all data specs specific for a given resource type into a single object which can be passed
    to the DataFetcher. It includes aliases for all specs so that they may be referenced by a user-defined
    shorthand and abstracted away from specific resources.

    Args:
        time_series (Dict[str, TimeSeriesSpec]): A dictionary mapping aliases to TimeSeriesSpecs.
        files (Dict[str, FileSpec]): A dicionary mapping aliases to FileSpecs.
    """

    def __init__(self, time_series: Dict[str, TimeSeriesSpec] = None, files: Dict[str, FileSpec] = None):
        self.time_series = time_series or {}
        self.files = files or {}
        self.validate()


class ScheduleInputTimeSeriesSpec(_BaseSpec):
    """Creates a ScheduleOutputTimeSeriesSpec.

    This object defines the time series a schedule should read from.

    If the granularity and aggregate parameters are omitted, the spec specifies raw data.

    Args:
        id (int): The id of the output time series.
        aggregate (str, optional): The aggregate function to apply to the time series.
        granularity (str, optional): Granularity of the datapoints. e.g. "1m", "2h", or "3d".
        include_outside_points (bool, optional): Whether or not to include the first point before and after start and
            end. Can only be used with raw data.
    """

    def __init__(self, id: int, aggregate: str = None, granularity: str = None, include_outside_points: bool = None):
        self.id = id
        self.aggregate = aggregate
        self.granularity = granularity
        self.include_outside_points = include_outside_points


class ScheduleInputSpec(_BaseSpec):
    """Creates a ScheduleInputSpec.

    The provided aliases must be the same as the input fields defined on the model.

    Args:
        time_series (Dict[str, ScheduleInputTimeSeriesSpec]): A dictionary mapping aliases to
            ScheduleInputTimeSeriesSpec objects.
    """

    def __init__(self, time_series: Dict[str, ScheduleInputTimeSeriesSpec] = None):
        self.time_series = time_series or {}


class ScheduleOutputTimeSeriesSpec(_BaseSpec):
    """Creates a ScheduleOutputTimeSeriesSpec.

    This object defines the time series a schedule should write to. You need to specify an offset which
    defines where in time your schedule can write data to for a given window. Offset defaults to 0, meaning that your
    schedule can write to the same time window which it was feeded data from.

    Args:
        id (int): The id of the output time series.
        offset (Union[int, str, timedelta], optional): The offset of the window to which your schedule is allowed to
            write data.
    """

    def __init__(self, id: int, offset: Union[int, str, timedelta] = 0):
        self.id = id
        self.offset = offset


class ScheduleOutputSpec(_BaseSpec):
    """Creates a ScheduleOutputSpec.

    The provided aliases must be the same as the output fields defined on the model.

    Args:
        time_series (Dict[str, ScheduleInputTimeSeriesSpec]): A dictionary mapping aliases to
            ScheduleOutputTimeSeriesSpec objects.
    """

    def __init__(self, time_series: Dict[str, ScheduleOutputTimeSeriesSpec] = None):
        self.time_series = time_series


class ScheduleDataSpec(_BaseSpec):
    """Creates a ScheduleDataSpec.

    This spec defines the input and output data for a given schedule, as well as how the hosting environment should
    feed the specified data to your model. This is done by specifying window size, a stride, and start time
    for the schedule.

    Args:
        input (ScheduleInputSpec): A schedule input spec describing input for a model.
        output (ScheduleOutputSpec): A schedule output spec describing output for a model.
        stride (Union[int, str, timedelta]): The interval at which predictions will be made. Can be either
            milliseconds, a timedelta object, or a time-string (e.g. "1h", "10d", "120s").
        window_size (Union[int, str, timedelta]): The size of each prediction window, i.e. how long back in time a
            prediction will look. Same format as stride.
        start (Union[int, str, datetime]): When the first prediction will be made.
    """

    def __init__(
        self,
        input: ScheduleInputSpec,
        output: ScheduleOutputSpec,
        stride: Union[int, str, timedelta],
        window_size: Union[int, str, timedelta],
        start: Union[int, str, datetime] = "now",
    ):
        self.input = input
        self.output = output
        self.stride = time_interval_to_ms(stride)
        self.window_size = time_interval_to_ms(window_size)
        self.start = timestamp_to_ms(start)

        self.validate()

    def get_instances(self, start: Union[int, str, datetime], end: Union[int, str, datetime]) -> List[DataSpec]:
        """Returns the DataSpec objects describing the prediction windows executed between start and end.

        Args:
            start (Union[str, int, datetime]): The start of the time period. Can be either milliseconds since epoch,
                time-ago format (e.g. "1d-ago"), or a datetime object.
            end (Union[str, int, datetime]): The end of the time period. Same format as start. Can also be set to "now".

        Returns:
            List[DataSpec]: List of DataSpec objects, one for each prediction window.
        """
        start, end = timestamp_to_ms(start), timestamp_to_ms(end)

        windows = calculate_windows(
            start=start, end=end, stride=self.stride, window_size=self.window_size, first=self.start
        )

        data_specs = []
        for start, end in windows:
            time_series_specs = {
                alias: TimeSeriesSpec(
                    id=spec.id,
                    start=start,
                    end=end,
                    aggregate=spec.aggregate,
                    granularity=spec.granularity,
                    include_outside_points=spec.include_outside_points,
                )
                for alias, spec in self.input.time_series.items()
            }
            data_specs.append(DataSpec(time_series=time_series_specs))
        return data_specs

    def get_execution_timestamps(self, start: Union[int, str, datetime], end: Union[int, str, datetime]) -> List[int]:
        """Returns a list of timestamps indicating when each prediction will be executed.

        This corresponds to the end of each DataSpec returned from get_instances().

        Args:
            start (Union[str, int, datetime]): The start of the time period. Can be either milliseconds since epoch,
                    time-ago format (e.g. "1d-ago"), or a datetime object.
            end (Union[str, int, datetime]): The end of the time period. Same format as start. Can also be set to "now".

        Returns:
            List[int]: A list of timestamps.
        """
        start, end = timestamp_to_ms(start), timestamp_to_ms(end)

        windows = calculate_windows(
            start=start, end=end, stride=self.stride, window_size=self.window_size, first=self.start
        )

        return [w[1] for w in windows]


class _BaseSchema(Schema):
    _ignore_values = [None, {}]
    _default_spec = None

    def __init__(self, *args, **kwargs):
        self._spec = kwargs.get("spec", None) or self._default_spec
        assert self._spec is not None
        if "spec" in kwargs:
            kwargs.pop("spec")
        super().__init__(*args, **kwargs)

    class Meta:
        unknown = RAISE

    @post_dump
    def remove_none(self, data):
        return {k: v for k, v in data.items() if v not in self._ignore_values}

    @post_load
    def to_spec(self, data):
        return self._spec(**data)


class AliasField(fields.String):
    def __init__(self):
        super().__init__(required=True, validate=self.__validate)

    def __validate(self, field_name):
        pattern = "[a-z]([a-z0-9_]{0,48}[a-z0-9])?"
        if not re.fullmatch(pattern, field_name):
            raise ValidationError(
                "Invalid alias. Must be 1 to 50 lowercase alphanumeric characters or `_`. Must start with a letter "
                "and cannot end with `_` ."
            )


class _TimeSeriesSpecSchema(_BaseSchema):
    _default_spec = TimeSeriesSpec

    id = fields.Int(required=True)
    start = fields.Int(required=True)
    end = fields.Int(required=True)
    aggregate = fields.Str(
        validate=validate.NoneOf(
            INVALID_AGGREGATE_FUNCTIONS, error="Not a valid aggregate function. Cannot use shorthand name."
        )
    )
    granularity = fields.Str()
    includeOutsidePoints = fields.Bool(attribute="include_outside_points")

    @validates_schema(skip_on_field_errors=False)
    def validate_aggregate(self, data):
        errors = {}

        if "aggregate" in data:
            if "granularity" not in data:
                errors["granularity"] = ["granularity must be specified for aggregates."]
            if "include_outside_points" in data and data["include_outside_points"]:
                errors["includeOutsidePoints"] = ["Can't include outside points for aggregates."]
        else:
            if "granularity" in data:
                errors["granularity"] = ["granularity can only be specified for aggregates."]

        if errors:
            raise ValidationError(errors)

    @validates("granularity")
    def validate_granularity(self, granularity):
        try:
            granularity_to_ms(granularity)
        except ValueError as e:
            raise ValidationError(str(e)) from e


class _FileSpecSchema(_BaseSchema):
    _default_spec = FileSpec

    id = fields.Int(required=True)


class _DataSpecSchema(_BaseSchema):
    _default_spec = DataSpec

    timeSeries = fields.Dict(keys=AliasField(), values=fields.Nested(_TimeSeriesSpecSchema), attribute="time_series")
    files = fields.Dict(keys=AliasField(), values=fields.Nested(_FileSpecSchema))


class _ScheduleInputDataSpecSchema(_BaseSchema):
    _default_spec = ScheduleInputSpec

    timeSeries = fields.Dict(
        keys=AliasField(),
        values=fields.Nested(_TimeSeriesSpecSchema(spec=ScheduleInputTimeSeriesSpec, exclude=("start", "end"))),
        attribute="time_series",
    )


class _ScheduleOutputTimeSeriesSpecSchema(_BaseSchema):
    _default_spec = ScheduleOutputTimeSeriesSpec

    id = fields.Int(required=True)
    offset = fields.Int(required=True)


class _ScheduleOutputDataSpecSchema(_BaseSchema):
    _default_spec = ScheduleOutputSpec

    timeSeries = fields.Dict(
        keys=AliasField(), values=fields.Nested(_ScheduleOutputTimeSeriesSpecSchema), attribute="time_series"
    )


class _ScheduleDataSpecSchema(_BaseSchema):
    _ignore_values = []
    _default_spec = ScheduleDataSpec

    input = fields.Nested(_ScheduleInputDataSpecSchema, required=True)
    output = fields.Nested(_ScheduleOutputDataSpecSchema, required=True)
    stride = fields.Int(required=True, validate=validate.Range(min=1))
    windowSize = fields.Int(required=True, attribute="window_size", validate=validate.Range(min=1))
    start = fields.Int(required=True, validate=validate.Range(min=0))


TimeSeriesSpec._schema = _TimeSeriesSpecSchema()
ScheduleInputTimeSeriesSpec._schema = _TimeSeriesSpecSchema(spec=ScheduleInputTimeSeriesSpec, exclude=("start", "end"))
FileSpec._schema = _FileSpecSchema()
DataSpec._schema = _DataSpecSchema()
ScheduleInputSpec._schema = _ScheduleInputDataSpecSchema()
ScheduleOutputTimeSeriesSpec._schema = _ScheduleOutputTimeSeriesSpecSchema()
ScheduleOutputSpec._schema = _ScheduleOutputDataSpecSchema()
ScheduleDataSpec._schema = _ScheduleDataSpecSchema()



import json


class DataFetcherHttpError(Exception):
    """Raised if an HTTP Error occurred while processing your request.

    Args:
        message (str):  The error message produced by the API.
        code (int):     The error code produced by the failure.
        x_request_id (str): The request-id generated for the failed request.
        extra (Dict):   A dict of any additional information.
    """

    def __init__(self, message, code=None, x_request_id=None, extra=None):
        self.message = message
        self.code = code
        self.x_request_id = x_request_id
        self.extra = extra

    def __str__(self):
        if self.extra:
            pretty_extra = json.dumps(self.extra, indent=4, sort_keys=True)
            return "{} | code: {} | X-Request-ID: {}\n{}".format(
                self.message, self.code, self.x_request_id, pretty_extra
            )
        return "{} | code: {} | X-Request-ID: {}".format(self.message, self.code, self.x_request_id)


class ApiKeyError(Exception):
    """Raised if the provided API key is missing or invalid."""

    pass


class DirectoryDoesNotExist(Exception):
    """Raised if the specified directory does not exist."""

    def __init__(self, directory):
        self.directory = directory

    def __str__(self):
        return "'{}' is not a directory".format(self.directory)


class InvalidFetchRequest(Exception):
    """Raised if an invalid fetch request is issued.

    For example if a request is issued for a time-aligned dataframe where the specified starts/ends or granularities
    of the time series are not the same.
    """

    pass


class InvalidAlias(Exception):
    """Raised if an invalid alias is specified."""

    def __init__(self, alias):
        self.alias = alias

    def __str__(self):
        return "Alias '{}' does not exist".format(self.alias)



import re
import time
from datetime import datetime, timedelta, timezone
from typing import List, Tuple, Union


def calculate_windows(start: int, end: int, stride: int, window_size: int, first: int) -> List[Tuple[int, int]]:
    next = max(start, first)
    if (next - first) % stride != 0:
        next += stride - ((start - first) % stride)
    windows = []
    while next < end:
        windows.append((next - window_size, next))
        next += stride
    return windows


_unit_in_ms_without_week = {"s": 1000, "m": 60000, "h": 3600000, "d": 86400000}
_unit_in_ms = {**_unit_in_ms_without_week, "w": 604800000}


def _time_string_to_ms(pattern, string, unit_in_ms):
    pattern = pattern.format("|".join(unit_in_ms))
    res = re.fullmatch(pattern, string)
    if res:
        magnitude = int(res.group(1))
        unit = res.group(2)
        return magnitude * unit_in_ms[unit]
    return None


def granularity_to_ms(granularity: str) -> int:
    """Returns millisecond representation of time-ago string"""
    ms = _time_string_to_ms(r"(\d+)({})", granularity, _unit_in_ms_without_week)
    if ms is None:
        raise ValueError(
            "Invalid granularity format: `{}`. Must be on format <integer>(s|m|h|d). E.g. '5m', '3h' or '1d'.".format(
                granularity
            )
        )
    return ms


def _time_ago_to_ms(time_ago_string: str) -> int:
    """Returns millisecond representation of time-ago string"""
    if time_ago_string == "now":
        return 0
    ms = _time_string_to_ms(r"(\d+)({})-ago", time_ago_string, _unit_in_ms)
    if ms is None:
        raise ValueError(
            "Invalid time-ago format: `{}`. Must be on format <integer>(s|m|h|d|w)-ago or 'now'. E.g. '3d-ago' or '1w-ago'.".format(
                time_ago_string
            )
        )
    return ms


def _datetime_to_ms(dt):
    return int(dt.replace(tzinfo=timezone.utc).timestamp() * 1000)


class NowCache:
    _cached_now = 0

    @classmethod
    def get_time_now(cls):
        now = int(round(time.time() * 1000))
        if now - cls._cached_now > 100:
            cls._cached_now = now
        return cls._cached_now


def timestamp_to_ms(t: Union[int, str, datetime]):
    """Returns the ms representation of some timestamp given by milliseconds, time-ago format or datetime object"""
    time_now = NowCache.get_time_now()
    if isinstance(t, int):
        ms = t
    elif isinstance(t, str):
        ms = time_now - _time_ago_to_ms(t)
    elif isinstance(t, datetime):
        ms = _datetime_to_ms(t)
    else:
        raise TypeError("Timestamp `{}` was of type {}, but must be int, str or datetime,".format(t, type(t)))

    if ms < 0:
        raise ValueError(
            "Timestamps can't be negative - they must represent a time after 1.1.1970, but {} was provided".format(ms)
        )

    return ms


def _time_interval_str_to_ms(interval_str: str):
    ms = _time_string_to_ms(r"(\d+)({})", interval_str, _unit_in_ms)
    if ms is None:
        raise ValueError(
            "Invalid time interval format: `{}`. Must be on format <integer>(s|m|h|d|w). E.g. '5m', '3h' or '1d'.".format(
                interval_str
            )
        )
    return ms


def time_interval_to_ms(t: Union[int, str, timedelta]):
    """Returns millisecond representation of time interval"""
    if isinstance(t, int):
        ms = t
    elif isinstance(t, str):
        ms = _time_interval_str_to_ms(t)
    elif isinstance(t, timedelta):
        ms = int(round(t.total_seconds() * 1000))
    else:
        raise TypeError("Time interval `{}` was of type {}, but must be int, str or timedelta,".format(t, type(t)))

    if ms <= 0:
        raise ValueError("Time interval has to be positive, but got {} ms".format(ms))

    return ms


def _time_offset_str_to_ms(offset_str: str):
    ms = _time_string_to_ms(r"(-?\d+)({})", offset_str, _unit_in_ms)
    if ms is None:
        raise ValueError(
            "Invalid time offset format: `{}`. Must be on format [-]<integer>(s|m|h|d|w). E.g. '-5m', '-3h' or '1d'.".format(
                offset_str
            )
        )
    return ms


def time_offset_to_ms(t: Union[int, str, timedelta]):
    if isinstance(t, int):
        return t
    elif isinstance(t, str):
        return _time_offset_str_to_ms(t)
    elif isinstance(t, timedelta):
        return int(round(t.total_seconds() * 1000))
    else:
        raise TypeError("Time offset `{}` was of type {}, but must be int, str or timedelta,".format(t, type(t)))



import re

from setuptools import find_packages, setup

packages = find_packages(exclude=["tests*"])

version = re.search('^__version__\s*=\s*"(.*)"', open("cognite/model_hosting/_version.py").read(), re.M).group(1)

setup(
    name="cognite-model-hosting",
    version=version,
    description="Utilities Cognite's model hosting environment",
    url="",  # TODO
    author="Nils Barlaug",
    author_email="nils.barlaug@cognite.com",
    packages=packages,
    install_requires=["pandas", "marshmallow==3.0.0rc4", "requests>=2.21.0,<3.0.0"],
    python_requires=">=3.5",
)



import json
from collections import defaultdict
from typing import Dict, List, Union

import pandas as pd
from marshmallow import RAISE, Schema, ValidationError, fields, validate

from cognite.model_hosting.schedules.exceptions import (
    DataframeMissingTimestampColumn,
    DuplicateAliasInScheduledOutput,
    InvalidScheduleOutputFormat,
)


def to_output(dataframe: Union[pd.DataFrame, List[pd.DataFrame]]):
    """Converts your data to a json serializable output format complying with the schedules feature.

    Args:
        dataframe (Union[List[pd.DataFrame, pd.DataFrame]]: A dataframe or list of dataframes.
    Returns:
        Dict: The data on a json serializable and schedules compliant output format.

    Examples:

        The correct output format looks like this::

            {
                "timeSeries":
                    {
                        "my-alias-1": [(t0, p0), (t1, p1), ...],
                        "my-alias-2": [(t0, p0), (t1, p1), ...],
                    }
            }
    """
    output = defaultdict(lambda: {})
    if isinstance(dataframe, pd.DataFrame):
        output["timeSeries"] = _convert_df_to_output_format(dataframe)
    elif isinstance(dataframe, List):
        for df in dataframe:
            if set(df.columns) - set(output["timeSeries"].keys()) != set(df.columns):
                raise DuplicateAliasInScheduledOutput("An alias has been provided multiple times")
            output["timeSeries"].update(_convert_df_to_output_format(df))
    else:
        raise TypeError("dataframe should be a pandas DataFrame or list of pandas DataFrames")
    return output


def _convert_df_to_output_format(df: pd.DataFrame):
    if not "timestamp" in df.columns:
        raise DataframeMissingTimestampColumn("DataFrame missing timestamp column")
    column_names = [col for col in df.columns if col != "timestamp"]
    output = {name: df[["timestamp", name]].values.tolist() for name in column_names}
    return output


class _ScheduleOutputSchema(Schema):
    class Meta:
        unknown = RAISE

    timeSeries = fields.Dict(
        keys=fields.Str(), values=fields.List(fields.List(fields.Float(), validate=validate.Length(equal=2)))
    )


_schedule_output_schema = _ScheduleOutputSchema(unknown=RAISE)


class ScheduleOutput:
    """Helper class for parsing and converting output from scheduled predictions.

    Args:
        output(Dict): The output returned from the scheduled prediction.
    """

    def __init__(self, output: Dict):
        self._output = self._load(output)

    def __str__(self):
        return json.dumps(self._output, indent=4, sort_keys=True)

    @staticmethod
    def _load(output):
        try:
            return _schedule_output_schema.load(output)
        except ValidationError as e:
            raise InvalidScheduleOutputFormat(e.messages) from e

    def _validate_alias(self, type: str, alias: str):
        assert self._output.get(type, {}).get(alias) is not None, "{} is not a valid alias".format(alias)

    def _validate_aligned(self, aliases: List[str]):
        timestamps = set()
        for alias in aliases:
            self._validate_alias("timeSeries", alias)
            timestamps.add(tuple(point[0] for point in self._output["timeSeries"][alias]))
        assert 1 == len(timestamps), "Timestamps for aliases {} are not aligned".format(aliases)

    def _get_dataframe_single_alias(self, alias) -> pd.DataFrame:
        self._validate_alias("timeSeries", alias)
        data = self._output["timeSeries"][alias]
        timestamps = [point[0] for point in data]
        datapoints = [point[1] for point in data]
        return pd.DataFrame({"timestamp": timestamps, alias: datapoints})

    def _get_dataframe_multiple_aliases(self, aliases: List[str]) -> pd.DataFrame:
        self._validate_aligned(aliases)
        data = {}
        timestamps = [p[0] for p in self._output["timeSeries"][aliases[0]]]
        data["timestamp"] = timestamps
        for a in aliases:
            data[a] = [p[1] for p in self._output["timeSeries"][a]]
        return pd.DataFrame(data)

    def get_dataframe(self, alias: Union[str, List[str]]) -> pd.DataFrame:
        """Returns a time-aligned dataframe of the specified alias(es).

        Assumes that all aliases specify output time series with matching timestamps.

        Args:
             alias(Union[str, List[str]]): alias or list of aliases

        Returns:
            pd.DataFrame: The dataframe containing the time series for the specified alias(es).
        """
        if isinstance(alias, str):
            return self._get_dataframe_single_alias(alias)
        elif isinstance(alias, List):
            return self._get_dataframe_multiple_aliases(alias)
        raise TypeError("alias must be a string or list of strings")

    def get_datapoints(self, alias: Union[str, List[str]]) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Returns the dataframes for the specified alias(es).

        Args:
            alias (Union[str, List[str]]): alias or list of aliases.

        Returns:
            Union[pd.DataFrame, Dict[str, pd.DataFrame]: A single dataframe if a single alias has been specified. Or a
                dictionary mapping alias to dataframe if a list of aliases has been provided.
        """
        if isinstance(alias, str):
            return self._get_dataframe_single_alias(alias)
        elif isinstance(alias, List):
            dataframes = {}
            for a in alias:
                dataframes[a] = self._get_dataframe_single_alias(a)
            return dataframes
        raise TypeError("alias must be a string or list of strings")



import os
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Dict, List, Union

import pandas as pd

from cognite.model_hosting.data_fetcher._client.cdp_client import CdpClient
from cognite.model_hosting.data_fetcher.exceptions import DirectoryDoesNotExist, InvalidAlias, InvalidFetchRequest
from cognite.model_hosting.data_spec import DataSpec, FileSpec, TimeSeriesSpec
from cognite.model_hosting.data_spec.exceptions import SpecValidationError

NUMBER_OF_THREADS = 16


def _execute_tasks_concurrently(func, tasks):
    with ThreadPoolExecutor(NUMBER_OF_THREADS) as p:
        futures = [p.submit(func, *task) for task in tasks]
        return [future.result() for future in futures]


class FileFetcher:
    """An object used for fetching files from CDP.

    .. attention:: This class should never be instantiated directly, but rather accessed through the DataFetcher class.

    Examples:
        Using the FileFetcher::

            from cognite.model_hosting.data_fetcher import DataFetcher

            data_fetcher = DataFetcher(data_spec=...)

            my_file = data_fetcher.files.fetch_to_memory(alias="my_file")
    """

    def __init__(self, file_specs: Dict[str, FileSpec], cdp_client: CdpClient):
        self._file_specs = file_specs
        self._cdp_client = cdp_client

    @property
    def aliases(self) -> List[str]:
        """Returns the file aliases defined in the data spec passed to the data fetcher.

        Returns:
            List[str]: The file aliases defined in the data spec passed to the data fetcher.
        """
        return list(self._file_specs.keys())

    def get_spec(self, alias: str) -> FileSpec:
        """Returns the FileSpec given by the alias

        Args:
            alias (str): The alias of the file.

        Returns:
            FileSpec: The file spec given by the alias.
        """
        if alias not in self.aliases:
            raise InvalidAlias(alias)
        return self._file_specs[alias].copy()

    def fetch(self, alias: Union[str, List[str]], directory: str = None) -> None:
        """Fetches the file(s) given by the provided alias(es) to a given directory.

        If provided, the directory must exist. If not provided, it will default to the current working directory.

        If a single alias is passed, a pandas DataFrame will be returned. If a list of aliases is passed, a dictionary
        which maps aliases to DataFrames is returned.

        Args:
            alias (Union[List[str], str]): The alias(es) to download file(s) for.
            directory(str, optional): The directory to download the file(s) to.

        Returns:
            None
        """
        directory = directory or os.getcwd()
        if not os.path.isdir(directory):
            raise DirectoryDoesNotExist(directory)

        if isinstance(alias, str):
            tasks = [(alias, directory)]
        elif isinstance(alias, list):
            tasks = [(a, directory) for a in alias]
        else:
            raise TypeError("alias must be of type str or list, was {}".format(type(alias)))

        _execute_tasks_concurrently(self._download_single_file, tasks)

    def fetch_to_memory(self, alias: Union[str, List[str]]) -> Union[bytes, Dict[str, bytes]]:
        """Fetches the file(s) given by the provided alias(es) to memory.

        If a list of aliases is passed, this method will return a dictionary mapping aliases to their respective file
        bytes.

        Args:
            alias (Union[List[str], str]): The alias(es) to download file(s) for.

        Returns:
            Union[bytes, Dict[str, bytes]]: The file(s).
        """
        if isinstance(alias, str):
            return self._download_single_file_to_memory(alias)[alias]
        elif isinstance(alias, list):
            files = {}
            tasks = [(a,) for a in alias]
            res = _execute_tasks_concurrently(self._download_single_file_to_memory, tasks)
            for file in res:
                files.update(file)
            return files
        raise TypeError("alias must be of type str or list, was {}".format(type(alias)))

    def _download_single_file(self, alias: str, directory: str):
        file_id = self.get_spec(alias).id
        file_path = os.path.join(directory, alias)
        self._cdp_client.download_file(file_id, file_path)

    def _download_single_file_to_memory(self, alias):
        file_id = self.get_spec(alias).id
        return {alias: self._cdp_client.download_file_to_memory(file_id)}


class TimeSeriesFetcher:
    """An object used for fetching time series data from CDP.

    .. attention:: This class should never be instantiated directly, but rather accessed through the DataFetcher class.

    Examples:
        Using the TimeSeriesFetcher::

            from cognite.model_hosting.data_fetcher import DataFetcher

            data_fetcher = DataFetcher(data_spec=...)

            my_datapoints = data_fetcher.time_series.fetch_datapoints(alias="my_ts_1")
    """

    def __init__(self, time_series_specs: Dict[str, TimeSeriesSpec], cdp_client: CdpClient):
        self._specs = time_series_specs
        self._cdp_client = cdp_client

    @property
    def aliases(self) -> List:
        """Returns the time series aliases defined in the data spec passed to the data fetcher.

        Returns:
            List[str]: The time series aliases defined in the data spec passed to the data fetcher.
        """
        return list(self._specs.keys())

    def get_spec(self, alias: str) -> TimeSeriesSpec:
        """Returns the TimeSeriesSpec given by the alias.

        Args:
            alias (str): The alias of the time series.

        Returns:
            TimeSeriesSpec: The time series spec given by the alias.
        """
        self._check_valid_alias(alias)
        return self._specs[alias].copy()

    def _check_valid_alias(self, alias: str):
        if type(alias) != str:
            raise TypeError("Alias `{}` ({}) was not a string".format(alias, type(alias)))
        if alias not in self._specs:
            raise InvalidAlias("No alias `{}` in the spec".format(alias))

    def _check_valid_aliases(self, aliases: List[str]):
        for alias in aliases:
            self._check_valid_alias(alias)

    def _check_only_aggregates(self, aliases: List[str]):
        for alias in aliases:
            if self._specs[alias].aggregate is None:
                raise InvalidFetchRequest("All time series of a data frame needs to be aggregates")

    def _get_common_start_end_granularity(self, aliases):
        starts = set()
        ends = set()
        granularities = set()
        for alias in aliases:
            starts.add(self._specs[alias].start)
            ends.add(self._specs[alias].end)
            granularities.add(self._specs[alias].granularity)

        if len(starts) != 1 or len(ends) != 1:
            raise InvalidFetchRequest(
                "The time series are not aligned. They need to have same start and end to be part of the same data frame"
            )

        if len(granularities) != 1:
            raise InvalidFetchRequest(
                "Granularity mismatch. All time series must have same granularity to be part of the same data frame"
            )

        return starts.pop(), ends.pop(), granularities.pop()

    def __convert_ts_names_to_aliases(self, df: pd.DataFrame) -> pd.DataFrame:
        name_to_label = {}
        ts_ids = [ts.id for ts in self._specs.values()]
        time_series = self._cdp_client.get_time_series_by_id(ts_ids)
        ts_names = {ts["id"]: ts["name"] for ts in time_series}
        for alias, ts_spec in self._specs.items():
            if ts_spec.aggregate:
                name_to_label[ts_names[ts_spec.id] + "|" + ts_spec.aggregate] = alias
            else:
                name_to_label[ts_names[ts_spec.id]] = alias
        return df.rename(columns=name_to_label)

    def fetch_dataframe(self, aliases: List[str]) -> pd.DataFrame:
        """Fetches a time-aligned dataframe of the time series specified by the provided aliases.

        This method requires that all specified aliases must refer to time series aggregates with the same granularity,
        start, and end.

        Args:
            aliases (List[str]): The list of aliases to retrieve a dataframe for.

        Returns:
            pandas.DataFrame: A pandas dataframe with the requested data.
        """
        if type(aliases) != list:
            raise TypeError("Invalid argument type. Aliases should be a list of string")
        self._check_valid_aliases(aliases)
        self._check_only_aggregates(aliases)
        time_series = []
        for alias in aliases:
            spec = self._specs[alias]
            time_series.append({"id": spec.id, "aggregate": spec.aggregate})
        start, end, granularity = self._get_common_start_end_granularity(aliases)
        df = self._cdp_client.get_datapoints_frame(time_series, granularity, start, end)
        df_with_alias_columns = self.__convert_ts_names_to_aliases(df)
        return df_with_alias_columns

    def _fetch_datapoints_single(self, alias):
        self._check_valid_alias(alias)

        spec = self._specs[alias]
        return self._cdp_client.get_datapoints_frame_single(
            spec.id, spec.start, spec.end, spec.aggregate, spec.granularity, spec.include_outside_points
        )

    def _fetch_datapoints_multiple(self, aliases):
        self._check_valid_aliases(aliases)

        tasks = []
        for alias in aliases:
            spec = self._specs[alias]
            tasks.append((spec.id, spec.start, spec.end, spec.aggregate, spec.granularity, spec.include_outside_points))
        res = _execute_tasks_concurrently(self._cdp_client.get_datapoints_frame_single, tasks)
        data_frames = {alias: df for alias, df in zip(aliases, res)}
        return data_frames

    def fetch_datapoints(self, alias: Union[str, List[str]]) -> Union[pd.DataFrame, Dict[str, pd.DataFrame]]:
        """Fetches dataframes for the time series specified by the aliases.

        If a single alias is passed, a pandas DataFrame will be returned. If a list of aliases is passed, a dictionary
        which maps aliases to DataFrames is returned.

        Args:
            alias (Union[List[str], str]): The alias(es) to retrieve data for.

        Returns:
            Union[pd.DataFrame, Dict[str, pd.DataFrame]: The requested dataframe(s).
        """
        if type(alias) == str:
            return self._fetch_datapoints_single(alias)
        elif type(alias) == list:
            return self._fetch_datapoints_multiple(alias)
        else:
            raise TypeError(
                "Invalid argument type. Specify either a single alias (string) or a list of aliases (list of strings)"
            )


class DataFetcher:
    """Creates an instance of DataFetcher.

    Args:
        data_spec (DataSpec): The data spec which describes the desired data.
        api_key (str, optional): API key for authenticating against CDP. Defaults to the value of the environment
                                variable "COGNITE_API_KEY".
        project (str, optional): Project. Defaults to project of given API key.
        base_url (str, optional): Base url to send requests to. Defaults to "https://api.cognitedata.com".
    """

    def __init__(
        self, data_spec: Union[DataSpec, Dict, str], api_key: str = None, project: str = None, base_url: str = None
    ):
        self._data_spec = self._load_data_spec(data_spec)
        self._data_spec.validate()
        self._cdp_client = CdpClient(api_key=api_key, project=project, base_url=base_url)

        self._files_fetcher = FileFetcher(self._data_spec.files, self._cdp_client)
        self._time_series_fetcher = TimeSeriesFetcher(self._data_spec.time_series, self._cdp_client)

    def _load_data_spec(self, data_spec):
        if type(data_spec) == DataSpec:
            return data_spec.copy()
        elif type(data_spec) == dict:
            return DataSpec.load(data_spec)
        elif type(data_spec) == str:
            return DataSpec.from_json(data_spec)
        else:
            raise SpecValidationError("data_spec has to be of type DataSpec, dict or str (json).")

    def get_data_spec(self):
        """Returns a copy of the DataSpec passed to the DataFetcher.

        Returns:
            DataSpec: A copy of the DataSpec passed to the DataFetcher.
        """
        return self._data_spec.copy()

    @property
    def files(self) -> FileFetcher:
        """Returns an instance of FileFetcher.

        Returns:
            FileFetcher: An instance of FileFetcher.
        """
        return self._files_fetcher

    @property
    def time_series(self) -> TimeSeriesFetcher:
        """Returns an instance of TimeSeriesFetcher.

        Returns:
            TimeSeriesFetcher: An instance of TimeSeriesFetcher.
        """
        return self._time_series_fetcher
